import csv
import logging
import os
import warnings
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Tuple

import pandas as pd
from sqlite_utils import Database

from .nem_objects import Reading
from .nem_reader import NEMFile
from .split_days import make_set_interval, split_multiday_reads

log = logging.getLogger(__name__)


def nmis_in_file(file_name) -> Generator[Tuple[str, List[str]], None, None]:
    """Return list of NMIs in file"""
    nf = NEMFile(file_name, strict=False)
    nf.nem_data()
    for nmi, suffixes in nf.nmi_channels.items():
        yield nmi, suffixes


def get_data_frame(
    channels: List[str],
    nmi_readings: Dict[str, List[Reading]],
    split_days: bool = False,
    set_interval: Optional[int] = None,
) -> pd.DataFrame:
    """Get a Pandas DataFrame for NMI"""

    warnings.warn("nemreader.get_data_frame is deprecated`.", DeprecationWarning)

    if split_days or set_interval:
        # Split any readings that are >24 hours
        for ch in channels:
            nmi_readings[ch] = list(split_multiday_reads(nmi_readings[ch]))

    if set_interval:
        for ch in channels:
            nmi_readings[ch] = list(make_set_interval(nmi_readings[ch], set_interval))

    # Work out which channel has smallest intervals and use that as first index
    ch_rows = []
    for i, ch in enumerate(channels):
        num_rows = len(nmi_readings[ch])
        ch_rows.append((num_rows, ch))
    ch_rows.sort()
    first_ch = ch_rows[-1][1]

    d = {
        "t_start": [x.t_start for x in nmi_readings[first_ch]],
        "t_end": [x.t_end for x in nmi_readings[first_ch]],
        "quality_method": [x.quality_method for x in nmi_readings[first_ch]],
        "event_code": [x.event_code for x in nmi_readings[first_ch]],
        "event_desc": [x.event_desc for x in nmi_readings[first_ch]],
    }
    d[first_ch] = [x.read_value for x in nmi_readings[first_ch]]

    df = pd.DataFrame(data=d, index=d["t_start"])

    for ch in channels:
        if ch == first_ch:
            continue  # This channel already added
        index = [x.t_start for x in nmi_readings[ch]]
        values = [x.read_value for x in nmi_readings[ch]]
        ser = pd.Series(data=values, index=index, name=ch)
        df.loc[:, ch] = ser
    return df


def output_as_data_frames(
    file_name,
    split_days: bool = True,
    set_interval: Optional[int] = None,
    strict: bool = False,
) -> List[Tuple[str, pd.DataFrame]]:
    """Return list of data frames for each NMI"""
    nf = NEMFile(file_name, strict=strict)
    data_frames = []
    for nmi, nmi_df in nf.get_per_nmi_dfs(
        split_days=split_days, set_interval=set_interval
    ):
        nmi_df.rename(columns={"quality": "quality_method"}, inplace=True)
        data_frames.append((nmi, nmi_df))
    return data_frames


def output_as_csv(file_name, output_dir=".", set_interval: Optional[int] = None):
    """
    Transpose all channels and output a csv that is easier
    to read and do charting on

    :param file_name: The NEM file to process
    :param output_dir: Specify different output location
    :returns: The file that was created
    """

    output_dir = Path(output_dir)
    output_paths = []
    os.makedirs(output_dir, exist_ok=True)
    nf = NEMFile(file_name, strict=False)
    df = nf.get_pivot_data_frame(set_interval=set_interval)
    for nmi in nf.nmis:
        nmi_df = df[(df["nmi"] == nmi)]
        del nmi_df["nmi"]
        last_date = nmi_df.iloc[-1][1].strftime("%Y%m%d")
        output_file = f"{nmi}_{last_date}_transposed.csv"
        output_path = output_dir / output_file
        nmi_df.to_csv(output_path, index=False)
        output_paths.append(output_path)
    return output_paths


def save_to_csv(headings: List[str], rows: List[list], output_path):
    """save data to csv file"""
    with open(output_path, "w", newline="") as csvfile:
        cwriter = csv.writer(
            csvfile, delimiter=",", quotechar='"', quoting=csv.QUOTE_MINIMAL
        )
        cwriter.writerow(headings)
        for row in rows:
            cwriter.writerow(row)
    log.debug("Created %s", output_path)
    return output_path


def flatten_and_group_rows(
    nmi: str,
    nmi_transactions: Dict[str, list],
    nmi_readings: Dict[str, List[Reading]],
    date_format: str = "%Y%m%d",
) -> List[list]:
    """Create flattened list of NMI reading data"""

    channels = list(nmi_transactions.keys())

    # Datastream suffix starting with a number are Accumulated Metering Data (NEM13)
    # Ensure no reading exceeds 24 hours
    split_required = False
    for ch in channels:
        if ch[0].isdigit():
            split_required = True
    if split_required:
        nmi_readings = split_multiday_reads(nmi_readings)

    rows = []
    for ch in channels:
        date_totals = {}
        date_qualities = {}
        uom = ""
        sn = ""
        for read in nmi_readings[ch]:
            t_end = read.t_start
            val = read.read_value
            uom = read.uom  # Only last value will be saved
            sn = read.meter_serial_number
            quality = read.quality_method
            t_group = t_end.strftime(date_format)
            try:
                date_totals[t_group] += val
            except KeyError:
                date_totals[t_group] = val

            if t_group not in date_qualities.keys():
                date_qualities[t_group] = set()
            date_qualities[t_group].add(quality)

        for day in date_totals.keys():
            day_total = date_totals[day]
            qualities = list(date_qualities[day])
            day_quality = "".join(qualities)
            if len(day_quality) > 1:
                day_quality = "V"  # Multiple quality methods
            row: List[Any] = [nmi, sn, day, ch, day_total, uom, day_quality]
            rows.append(row)
    return rows


def output_as_daily_csv(file_name, output_dir="."):
    """
    Transpose all channels and output a daily csv that is easier
    to read and do charting on

    :param file_name: The NEM file to process
    :param output_dir: Specify different output location
    :returns: The file that was created
    """

    output_dir = Path(output_dir)
    os.makedirs(output_dir, exist_ok=True)

    file_stem = Path(file_name).stem
    output_file = f"{file_stem}_daily_totals.csv"
    output_path = output_dir / output_file

    nf = NEMFile(file_name, strict=False)
    m = nf.nem_data()
    nmis = m.readings.keys()
    all_rows = []
    headings = [
        "nmi",
        "meter_sn",
        "day",
        "channel",
        "day_total",
        "uom",
        "quality_method",
    ]
    for nmi in nmis:
        rows = flatten_and_group_rows(nmi, m.transactions[nmi], m.readings[nmi])
        for row in rows:
            all_rows.append(row)

    save_to_csv(headings, all_rows, output_path)

    return output_path


def output_as_sqlite(
    file_name: Path,
    output_dir=".",
    split_days: bool = False,
    set_interval: Optional[int] = None,
):
    """Export all channels to sqlite file"""

    output_dir = Path(output_dir)
    os.makedirs(output_dir, exist_ok=True)
    output_path = output_dir / "nemdata.db"
    db = Database(output_path)

    nf = NEMFile(file_name, strict=False)
    m = nf.nem_data()
    nmis = m.readings.keys()
    for nmi in nmis:
        channels = list(m.transactions[nmi].keys())
        nmi_readings = m.readings[nmi]

        for ch in channels:
            if split_days or set_interval:
                nmi_readings[ch] = list(split_multiday_reads(nmi_readings[ch]))

            if set_interval:
                nmi_readings[ch] = list(
                    make_set_interval(nmi_readings[ch], set_interval)
                )

            items = []
            for x in nmi_readings[ch]:
                item = {
                    "nmi": nmi,
                    "channel": ch,
                    "t_start": x.t_start,
                    "t_end": x.t_end,
                    "value": x.read_value,
                    "quality_method": x.quality_method,
                    "event_code": x.event_code,
                    "event_desc": x.event_desc,
                }
                items.append(item)
            db["readings"].upsert_all(
                items,
                pk=("nmi", "channel", "t_start"),
                column_order=("nmi", "channel", "t_start"),
            )

    db.create_view(
        "nmi_summary",
        """
        SELECT nmi, channel, MIN(t_start) as first_interval, MAX(t_end) as last_interval
        FROM readings
        GROUP BY nmi, channel
    """,
        replace=True,
    )
    return output_path



import logging
import os
from pathlib import Path
from typing import Optional

import typer

from .outputs import nmis_in_file, output_as_csv, output_as_daily_csv, output_as_sqlite
from .version import __version__

LOG_FORMAT = "%(asctime)s %(levelname)-8s %(message)s"
app = typer.Typer()
DEFAULT_DIR = Path(".")


def version_callback(value: bool):
    if value:
        typer.echo(f"nemreader version: {__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    version: Optional[bool] = typer.Option(
        None, "--version", callback=version_callback
    ),
):
    """nemreader

    Parse AEMO NEM12 and NEM13 meter data files
    """
    pass


@app.command()
def list_nmis(nemfile: Path, verbose: bool = typer.Option(False, "--verbose", "-v")):
    if verbose:
        log_level = "DEBUG"
    else:
        log_level = "WARNING"
    logging.basicConfig(level=log_level, format=LOG_FORMAT)

    nmis = list(nmis_in_file(nemfile))
    typer.echo("The following NMI[suffix] exist in this file:")
    for nmi, suffixes in nmis:
        suffix_str = ",".join(suffixes)
        typer.echo(f"{nmi}[{suffix_str}]")


@app.command()
def output_csv(
    nemfile: Path,
    verbose: bool = typer.Option(False, "--verbose", "-v"),
    set_interval: Optional[int] = None,
    outdir: Path = typer.Option(
        DEFAULT_DIR,
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
    ),
):
    """Output NEM file to transposed CSV.

    nemfile is the name of the file to parse.
    """
    if verbose:
        log_level = "DEBUG"
    else:
        log_level = "WARNING"
    logging.basicConfig(level=log_level, format=LOG_FORMAT)
    for fname in output_as_csv(nemfile, output_dir=outdir, set_interval=set_interval):
        typer.echo(f"Created {fname}")


@app.command()
def output_csv_daily(
    nemfile: Path,
    verbose: bool = typer.Option(False, "--verbose", "-v"),
    outdir: Path = typer.Option(
        DEFAULT_DIR,
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
    ),
):
    """Output NEM file to transposed CSV.

    nemfile is the name of the file to parse.
    """
    if verbose:
        log_level = "DEBUG"
    else:
        log_level = "WARNING"
    logging.basicConfig(level=log_level, format=LOG_FORMAT)
    fname = output_as_daily_csv(nemfile, output_dir=outdir)
    typer.echo(f"Created {fname}")


@app.command()
def output_sqlite(
    nemfile: Path,
    outdir: Path = typer.Option(
        DEFAULT_DIR,
        exists=True,
        file_okay=True,
        dir_okay=True,
        writable=True,
    ),
    verbose: bool = typer.Option(False, "--verbose", "-v"),
    five_min: bool = typer.Option(False, "--five-min", "-5"),
):
    """Output NEM file to transposed CSV.

    nemfile is the name of the file to parse.
    """
    if verbose:
        log_level = "DEBUG"
    else:
        log_level = "WARNING"
    logging.basicConfig(level=log_level, format=LOG_FORMAT)
    if os.path.isdir(nemfile):
        typer.echo(f"Getting files in directory {nemfile}")
        files = list(nemfile.glob("*.csv"))
        files += list(nemfile.glob("*.zip"))
    else:
        files = [nemfile]
    for fp in files:
        typer.echo(f"Processing {fp}")
        try:
            output_as_sqlite(fp, output_dir=outdir, make_fivemins=five_min)
        except Exception:
            typer.echo(f"Not a valid nem file: {fp}")
    typer.echo("Finished exporting to DB.")



from nemreader import (
    nmis_in_file,
    output_as_csv,
    output_as_daily_csv,
    output_as_data_frames,
)


def test_nmi_output(tmpdir):
    """Check NMIs in file"""
    file_name = "examples/unzipped/Example_NEM12_multiple_meters.csv"
    items = list(nmis_in_file(file_name))
    nmis = [x[0] for x in items]
    assert "NCDE001111" in nmis
    assert "NDDD001888" in nmis

    suffixes = [x[1] for x in items]
    assert "E1" in suffixes[0]
    assert "E1" not in suffixes[1]
    assert "B1" in suffixes[1]


def test_csv_output(tmpdir):
    """Create a temporary csv output"""
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"
    output_files = output_as_csv(file_name, output_dir=tmpdir)
    assert len(output_files) == 1


def test_daily_csv_output(tmpdir):
    """Create a temporary csv output"""
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"
    output_file = output_as_daily_csv(file_name, output_dir=tmpdir)
    assert "Example_NEM12_actual_interval_daily_totals.csv" in str(output_file)


def test_data_frame_output():
    """Create a pandas dataframe"""
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"
    output_dfs = output_as_data_frames(file_name)
    for nmi, df in output_dfs:
        assert type(nmi) == str
        assert df["quality_method"][0] == "A"


def test_data_frame_output_different_interval():
    """Create a pandas dataframe"""
    file_name = "examples/unzipped/Example_NEM12_different_interval_length.csv"
    output_dfs = output_as_data_frames(file_name)
    nmi, df = output_dfs[0]  # Return data for first NMI in file
    assert type(nmi) == str
    assert df["V1"].count() == 144
    assert df["E1"].count() == 48


def test_data_frame_output_nem13():
    """Create a pandas dataframe"""
    file_name = "examples/unzipped/Example_NEM13_forward_estimate.csv"
    output_dfs = output_as_data_frames(file_name)
    for nmi, df in output_dfs:
        print(df.head())
        assert type(nmi) == str
        assert df["quality_method"][0] == "E64"



from nemreader import read_nem_file


def test_correct_nmis_legacy():
    """Legacy way of reading in data"""
    meter_data = read_nem_file("examples/unzipped/Example_NEM12_actual_interval.csv")
    assert len(meter_data.readings) == 1
    assert "VABD000163" in meter_data.readings



from nemreader import NEMFile


def test_optional_scheduled_read():
    nf = NEMFile("examples/unzipped/Example_NEM12_no_scheduled_read.csv", strict=True)
    meter_data = nf.nem_data()
    readings = meter_data.readings["NMI111"]["E1"]
    assert len(readings) == 96



from datetime import datetime, timedelta
from statistics import mean
from typing import Generator, Iterable, Tuple

from .nem_objects import Reading


def split_multiday_reads(
    readings: Iterable[Reading],
) -> Generator[Reading, None, None]:
    """Split readings into daily intervals if they exceed 24 hours"""
    for r in readings:
        interval_s = int((r.t_end - r.t_start).total_seconds())
        interval_days = interval_s / 60 / 60 / 24
        if interval_days <= 1.0:
            # Don't need to do anything
            yield r
        else:
            for start, end, val in split_reading_into_days(
                r.t_start, r.t_end, r.read_value
            ):
                yield Reading(
                    start,
                    end,
                    val,
                    r.uom,
                    r.meter_serial_number,
                    r.quality_method,
                    r.event_code,
                    r.event_desc,
                    None,
                    None,
                )


def split_reading_into_days(start, end, val):
    """Split a single reading into even daily readings"""
    total_secs = (end - start).total_seconds()
    # Split first day into single day
    next_day = start.replace(hour=0, minute=0, second=0) + timedelta(days=1)
    fd_secs = (next_day - start).total_seconds()
    fd_val = val * (fd_secs / total_secs)
    yield start, next_day, fd_val

    # Generate the rest of the days
    period_start = next_day
    period_end = period_start + timedelta(days=1)
    if period_end >= end:
        period_end = end
    while period_start < end:
        if period_end > end:
            period_end = end
        period_secs = (period_end - period_start).total_seconds()
        period_val = val * (period_secs / total_secs)
        yield period_start, period_end, period_val,
        period_start += timedelta(days=1)
        period_end += timedelta(days=1)


def new_intervals(
    start_date: datetime, end_date: datetime, interval: float = 5
) -> Generator[Tuple[datetime, datetime], None, None]:
    """Generate equally spaced intervals between two dates"""
    delta = timedelta(seconds=interval * 60)
    orig_delta = end_date - start_date
    if (orig_delta / delta) % 1 != 0:
        raise ValueError(f"New interval of {delta} not an increment of {orig_delta}")
    num_intervals = int(orig_delta / delta)
    for i in range(0, num_intervals):
        start = start_date + i * delta
        end = start + delta
        yield start, end


def get_group_end(end_date: datetime, interval: int = 30) -> datetime:
    """Get interval group end time"""
    group_end = end_date
    while group_end.minute % interval != 0:
        group_end += timedelta(minutes=1)
    return group_end


def make_set_interval(
    readings: Iterable[Reading], new_interval: int = 5
) -> Iterable[Reading]:
    """Generate equally spaced values at 5-min intervals"""
    delta = timedelta(seconds=new_interval * 60)
    group_records = {}
    for r in readings:
        interval = r.t_end - r.t_start
        if interval == delta:  # No change required
            yield r
            continue

        if interval > delta:  # Need to split up smaller
            intervals = list(new_intervals(r.t_start, r.t_end, interval=new_interval))
            if r.uom and r.uom[-1].lower() == "h":
                split_val = r.read_value / len(intervals)
            else:
                split_val = r.read_value  # Average so assume flat

            for start, end in intervals:
                yield Reading(
                    start,
                    end,
                    split_val,
                    r.uom,
                    r.meter_serial_number,
                    r.quality_method,
                    r.event_code,
                    r.event_desc,
                    None,
                    None,
                )

        if interval < delta:  # Need to aggregate to bigger intervals
            # Increment dictionary value
            group_end = get_group_end(r.t_end, new_interval)
            if group_end not in group_records:
                group_records[group_end] = [r]
            else:
                group_records[group_end].append(r)

    # Output any aggregated grouped values
    for group_end in sorted(group_records.keys()):

        start = group_end - delta
        grp_readings = group_records[group_end]
        uom = grp_readings[0].uom
        meter_serial_number = grp_readings[0].meter_serial_number
        quality_methods = list(set([x.quality_method for x in grp_readings]))
        event_codes = list(set([x.event_code for x in grp_readings if x.event_code]))
        event_descs = list(set([x.event_desc for x in grp_readings if x.event_desc]))
        if len(quality_methods) == 1:
            quality_method = grp_readings[0].quality_method
            event_code = event_codes[0] if event_codes else ""
            event_desc = event_descs[0] if event_descs else ""
        else:
            quality_method = "V"
            event_code = grp_readings[0].event_code
            event_desc = grp_readings[0].event_desc

        if r.uom and r.uom[-1].lower() == "h":
            grp_value = sum([x.read_value for x in grp_readings])
        else:
            grp_value = mean([x.read_value for x in grp_readings])

        yield Reading(
            start,
            group_end,
            grp_value,
            uom,
            meter_serial_number,
            quality_method,
            event_code,
            event_desc,
            None,
            None,
        )



""" Test Suite
"""

import pytest

from nemreader import NEMFile


def test_nem13_readings():
    """Test the NEM13 consumption data"""
    nf = NEMFile("examples/unzipped/Example_NEM13_consumption_data.csv", strict=True)
    meter_data = nf.nem_data()
    readings = meter_data.readings["VABC005890"]["11"]
    unit_count = 0
    for record in readings:
        unit_count += record[2]  # Read value should be third record in tuple
        assert unit_count == pytest.approx(1312.1, 0.1)



"""
    nemreader
    ~~~~~
    Parse AEMO NEM12 (interval metering data) and
    NEM13 (accumulated metering data) data files
"""

import logging
from logging import NullHandler

from .nem_reader import NEMFile, read_nem_file
from .outputs import (
    nmis_in_file,
    output_as_csv,
    output_as_daily_csv,
    output_as_data_frames,
    output_as_sqlite,
)
from .version import __version__

__all__ = [
    "__version__",
    "NEMFile",
    "read_nem_file",
    "nmis_in_file",
    "output_as_csv",
    "output_as_daily_csv",
    "output_as_data_frames",
    "output_as_sqlite",
]

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(NullHandler())



import datetime

from nemreader import NEMFile


def test_Date12_parse():
    """test that Date12 is parsed correctly"""
    # 200402070911 = 2004-02-07, 09:11
    nf = NEMFile("examples/unzipped/Example_NEM12_multiple_meters.csv", strict=True)
    meter_data = nf.nem_data()
    assert meter_data.header.creation_date == datetime.datetime(2004, 2, 7, 9, 11)



""" Test Suite
"""

import pandas as pd

from nemreader.outputs import output_as_data_frames


def test_split_interval():
    """Create a temporary csv output"""
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"

    for _, df in output_as_data_frames(file_name, set_interval=1):
        first = df.iloc[0]
        delta = pd.Timedelta(first.t_end - first.t_start).seconds
        assert delta == 60 * 1

        assert first.E1 == 1.111 / 30


def test_group_interval():
    """Create a temporary csv output"""
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"

    for _, df in output_as_data_frames(file_name, set_interval=60):
        first = df.iloc[0]
        delta = pd.Timedelta(first.t_end - first.t_start).seconds
        assert delta == 60 * 60

        assert first.E1 == 1.111 * 2



import os

from nemreader import NEMFile


def test_unzipped_examples():
    """Open and parse unzipped example files"""

    skips = ["Example_NEM12_powercor.csv", "Example_NEM12_powercor_missing_fields.csv"]
    test_path = os.path.abspath("examples/unzipped")
    for file_name in os.listdir(test_path):
        if file_name in skips:
            continue
        test_file = os.path.join(test_path, file_name)
        nf = NEMFile(test_file, strict=True)
        meter_data = nf.nem_data()
        assert meter_data.header.version_header in ["NEM12", "NEM13"]


def test_nem12_examples():
    """Open and parse zipped NEM12 example files"""
    skips = [
        "NEM12#Scenario10#ETSAMDP#NEMMCO.zip",  # 300 Row has new line
        "Example_NEM12_powercor.csv.zip",
    ]
    test_path = os.path.abspath("examples/nem12")
    for file_name in os.listdir(test_path):
        if file_name in skips:
            continue
        test_file = os.path.join(test_path, file_name)
        nf = NEMFile(test_file, strict=True)
        meter_data = nf.nem_data()
        assert meter_data.header.version_header == "NEM12"


def test_nem13_examples():
    """Open and parse zipped NEM13 example files"""

    test_path = os.path.abspath("examples/nem13")
    for file_name in os.listdir(test_path):
        test_file = os.path.join(test_path, file_name)
        nf = NEMFile(test_file, strict=True)
        meter_data = nf.nem_data()
        assert meter_data.header.version_header == "NEM13"



import pytest
from typer.testing import CliRunner

from nemreader.cli import app


@pytest.fixture
def runner():
    return CliRunner()


def test_cli_list_nmis(runner):
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"
    result = runner.invoke(app, ["list-nmis", file_name, "--verbose"])
    assert "The following NMI[suffix] exist in this file:" in result.stdout
    assert result.exit_code == 0


def test_cli_csv(runner):
    file_name = "examples/invalid/Example_NEM12_missing_header.csv"
    result = runner.invoke(app, ["output-csv", file_name, "--verbose"])
    assert "Created" in result.stdout
    assert result.exit_code == 0


def test_cli_sqlite(runner):
    file_name = "examples/unzipped/Example_NEM12_actual_interval.csv"
    result = runner.invoke(app, ["output-sqlite", file_name, "--verbose"])
    assert "Finished exporting to DB." in result.stdout
    assert result.exit_code == 0



import pytest

from nemreader import NEMFile


def test_correct_NMIs():

    nf = NEMFile("examples/invalid/Example_NEM12_powercor.csv", strict=False)
    meter_data = nf.nem_data()
    assert len(meter_data.readings) == 1
    assert "VABD000163" in meter_data.readings


def test_incomplete_interval_row():
    nf = NEMFile("examples/invalid/Example_NEM12_incomplete_interval.csv", strict=False)
    meter_data = nf.nem_data()
    assert len(meter_data.readings) == 1
    assert "VABD000163" in meter_data.readings


def test_correct_channels():
    nf = NEMFile("examples/invalid/Example_NEM12_powercor.csv", strict=False)
    meter_data = nf.nem_data()
    readings = meter_data.readings["VABD000163"]
    assert len(readings) == 2
    assert "E1" in readings
    assert "Q1" in readings


def test_correct_records():
    nf = NEMFile("examples/invalid/Example_NEM12_powercor.csv", strict=False)
    meter_data = nf.nem_data()
    readings = meter_data.readings["VABD000163"]

    assert len(readings["E1"]) == 96
    assert readings["E1"][10].read_value == pytest.approx(1.11, 0.1)
    assert readings["E1"][-1].read_value == pytest.approx(3.33, 0.1)

    assert len(readings["Q1"]) == 96
    assert readings["Q1"][10].read_value == pytest.approx(2.22, 0.1)
    assert readings["Q1"][-1].read_value == pytest.approx(4.44, 0.1)
    assert readings["Q1"][-1].quality_method == "A"


def test_zipped_load():
    nf = NEMFile("examples/invalid/Example_NEM12_powercor.csv.zip", strict=False)
    meter_data = nf.nem_data()
    assert "VABD000163" in meter_data.nmis


def test_missing_fields():
    nf = NEMFile("examples/invalid/Example_NEM12_powercor.csv", strict=False)
    meter_data = nf.nem_data()
    assert "VABD000163" in meter_data.nmis



import pytest

from nemreader import NEMFile


def test_correct_records():
    nf = NEMFile("examples/unzipped/Example_NEM12_multiple_quality.csv", strict=True)
    meter_data = nf.nem_data()
    readings = meter_data.readings["CCCC123456"]["E1"]
    assert len(readings) == 48
    assert readings[0].read_value == pytest.approx(18.023, 0.1)
    assert readings[-1].read_value == pytest.approx(14.733, 0.1)


def test_correct_quality():
    nf = NEMFile("examples/unzipped/Example_NEM12_multiple_quality.csv", strict=True)
    meter_data = nf.nem_data()
    readings = meter_data.readings["CCCC123456"]["E1"]
    assert readings[0].quality_method == "F14"
    assert readings[10].quality_method == "F14"
    assert readings[19].quality_method == "F14"
    assert readings[20].quality_method == "A"
    assert readings[22].quality_method == "A"
    assert readings[23].quality_method == "A"
    assert readings[25].quality_method == "S14"
    assert readings[30].quality_method == "S14"
    assert readings[47].quality_method == "S14"



from datetime import datetime
from typing import Dict, List, NamedTuple, Optional

from pydantic import BaseModel


class HeaderRecord(NamedTuple):
    """Header record (100)"""

    version_header: str
    creation_date: Optional[datetime]
    from_participant: str
    to_participant: str
    file_name: str
    assumed: bool


class NmiDetails(NamedTuple):
    """NMI data details record (200)"""

    nmi: str
    nmi_configuration: str
    register_id: str
    nmi_suffix: str
    mdm_datastream_identifier: str
    meter_serial_number: str
    uom: str
    interval_length: int
    next_scheduled_read_date: Optional[datetime]


class Reading(NamedTuple):
    """Represents a meter reading"""

    t_start: datetime
    t_end: datetime
    read_value: float
    uom: str
    meter_serial_number: str
    quality_method: Optional[str]
    event_code: Optional[str]
    event_desc: Optional[str]
    # Below attributes relevant for NEM13 only
    val_start: Optional[float]
    val_end: Optional[float]


class BasicMeterData(NamedTuple):
    """Basic meter data record (250)"""

    nmi: str
    nmi_configuration: str
    register_id: str
    nmi_suffix: str
    mdm_data_stream_identifier: str
    meter_serial_number: str
    direction_indicator: str
    previous_register_read: str
    previous_register_read_datetime: Optional[datetime]
    previous_quality_method: str
    previous_reason_code: int
    previous_reason_description: str
    current_register_read: str
    current_register_read_datetime: Optional[datetime]
    current_quality_method: str
    current_reason_code: int
    current_reason_description: str
    quantity: float
    uom: str
    next_scheduled_read_date: Optional[datetime]
    update_datetime: Optional[datetime]
    msats_load_datetime: Optional[datetime]


class IntervalRecord(NamedTuple):
    """Interval data record (300)"""

    interval_date: Optional[datetime]
    interval_values: List[Reading]
    quality_method: str
    meter_serial_number: str
    reason_code: str
    reason_description: str
    update_datetime: Optional[datetime]
    msats_load_datatime: Optional[datetime]


class EventRecord(NamedTuple):
    """Interval event record (400)"""

    start_interval: int
    end_interval: int
    quality_method: str
    reason_code: str
    reason_description: str


class B2BDetails12(NamedTuple):
    """B2B details record (500)"""

    trans_code: str
    ret_service_order: str
    read_datetime: Optional[datetime]
    index_read: str


class B2BDetails13(NamedTuple):
    """B2B details record (550)"""

    previous_trans_code: str
    previous_ret_service_order: str
    current_trans_code: str
    current_ret_service_order: str


class NEMReadings(BaseModel):
    """Represents a meter reading"""

    readings: Dict[str, Dict[str, List[Reading]]]
    transactions: Dict[str, Dict[str, list]]


class NEMData(BaseModel):
    """Represents a meter reading"""

    header: HeaderRecord
    readings: Dict[str, Dict[str, List[Reading]]]
    transactions: Dict[str, Dict[str, list]]

    @property
    def nmis(self) -> List[str]:
        return list(self.transactions.keys())



import numpy as np
import pandas as pd

from nemreader import NEMFile


def test_unit_capitalisation():
    """Check that capitalisation of units does not affect the sum/aggregation
    of data when resampling the data."""

    nf = NEMFile("examples/unzipped/Example_NEM12_upper_case_units.csv", strict=True)
    meter_data = nf.nem_data()

    df_15min = nf.get_pivot_data_frame(split_days=True, set_interval=15)
    df_30min = nf.get_pivot_data_frame(split_days=True, set_interval=30)
    df_60min = nf.get_pivot_data_frame(split_days=True, set_interval=60)

    for (nmi, nmi_readings) in meter_data.readings.items():
        for key in nmi_readings.keys():
            assert nmi_readings[key][0].t_end - nmi_readings[key][
                0
            ].t_start == pd.Timedelta(minutes=30)
            assert "H" in nmi_readings[key][0].uom

        nmi_df_15min = df_15min[(df_15min["nmi"] == nmi)]
        nmi_df_30min = df_30min[(df_30min["nmi"] == nmi)]
        nmi_df_60min = df_60min[(df_60min["nmi"] == nmi)]
        # Units are in KWH - expect the values to be summed
        np.testing.assert_allclose(
            np.sum(nmi_df_30min["E1"]), np.sum(nmi_df_60min["E1"])
        )
        np.testing.assert_allclose(
            np.sum(nmi_df_30min["E1"]), np.sum(nmi_df_15min["E1"])
        )



import csv
import logging
import os
import zipfile
from datetime import datetime, timedelta
from itertools import chain, islice
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple

import pandas as pd

from .nem_objects import (
    B2BDetails12,
    B2BDetails13,
    BasicMeterData,
    EventRecord,
    HeaderRecord,
    IntervalRecord,
    NEMData,
    NEMReadings,
    NmiDetails,
    Reading,
)
from .split_days import make_set_interval, split_multiday_reads

log = logging.getLogger(__name__)


class NEMFile:
    """An NEM file object"""

    def __init__(self, file_path: str, strict: bool = False) -> None:

        self.file_path = file_path
        self.strict = strict
        self.nmis: set = set()
        self.nmi_channels: dict = {}

    def __repr__(self):
        return f"<NEMFile {self.file_path}>"

    @property
    def zipped(self) -> bool:
        """Check whether file is zipped or not"""
        _, file_extension = os.path.splitext(self.file_path)
        if file_extension.lower() == ".zip":
            return True
        return False

    def parse_nem_file(self, nem_file, file_name="") -> NEMReadings:
        """Parse NEM file and return meter readings named tuple"""
        reader = csv.reader(nem_file, delimiter=",")
        first_row = next(reader, None)

        # Some Powercor/Citipower files have empty line at start, skip if so.
        if not first_row:
            first_row = next(reader, None)

        try:
            record_indicator = int(first_row[0])
        except Exception:
            record_indicator = 0

        if not first_row or record_indicator != 100:
            if self.strict:
                raise ValueError("NEM Files must start with a 100 row")
            else:
                log.warning("Missing header (100) row, assuming NEM12.")
                header = HeaderRecord("NEM12", None, "", "", file_name, assumed=True)
        else:
            header = parse_100_row(first_row, file_name)
            if header.version_header not in ["NEM12", "NEM13"]:
                raise ValueError("Invalid NEM version {}".format(header.version_header))

        self.header = header
        if header.assumed:
            # We have to parse the first row again so we don't miss any data.
            reader = chain([first_row], reader)
            return parse_nem12_rows(reader, file_name=nem_file)
        if header.version_header == "NEM12":
            return parse_nem12_rows(reader, file_name=nem_file)
        else:
            return parse_nem13_rows(reader)

    def nem_data(self) -> NEMData:
        """Return data in legacy data format"""
        if self.zipped:
            log.debug("Extracting zip file")
            with zipfile.ZipFile(self.file_path, "r") as archive:
                for csv_file in archive.namelist():
                    with archive.open(csv_file) as csv_text:
                        # Zip file is open in binary mode
                        # So decode then convert back to list
                        nmi_file = csv_text.read().decode("utf-8").splitlines()
                        reads = self.parse_nem_file(nmi_file, file_name=csv_file)
                        for nmi in reads.transactions.keys():
                            self.nmis.add(nmi)
                            suffixes = list(reads.transactions[nmi].keys())
                            self.nmi_channels[nmi] = suffixes
                        return NEMData(
                            header=self.header,
                            readings=reads.readings,
                            transactions=reads.transactions,
                        )

        with open(self.file_path) as nmi_file:
            reads = self.parse_nem_file(nmi_file)
            for nmi in reads.transactions.keys():
                self.nmis.add(nmi)
                suffixes = list(reads.transactions[nmi].keys())
                self.nmi_channels[nmi] = suffixes
            return NEMData(
                header=self.header,
                readings=reads.readings,
                transactions=reads.transactions,
            )

    def get_data_frame(
        self, split_days: bool = False, set_interval: Optional[int] = None
    ) -> pd.DataFrame:
        """Return NEMData as a DataFrame"""
        nd = self.nem_data()
        frames = []
        for nmi in nd.readings.keys():
            for suffix in nd.readings[nmi].keys():
                reads = nd.readings[nmi][suffix]

                if split_days or set_interval:
                    reads = list(split_multiday_reads(reads))

                if set_interval:
                    reads = list(make_set_interval(reads, set_interval))

                data = {
                    "nmi": [nmi for _ in range(len(reads))],
                    "suffix": [suffix for _ in range(len(reads))],
                    "serno": [x.meter_serial_number for x in reads],
                    "t_start": [x.t_start for x in reads],
                    "t_end": [x.t_end for x in reads],
                    "value": [x.read_value for x in reads],
                    "quality": [x.quality_method for x in reads],
                    "evt_code": [x.event_code for x in reads],
                    "evt_desc": [x.event_desc for x in reads],
                }
                frames.append(pd.DataFrame(data))
        return pd.concat(frames)

    def get_pivot_data_frame(
        self,
        split_days: bool = False,
        set_interval: Optional[int] = None,
        include_serno: bool = False,
    ) -> pd.DataFrame:
        """Return NEMData as a DataFrame with suffix columns"""
        df = self.get_data_frame(split_days, set_interval)
        pivot_index = ["nmi", "t_start", "t_end", "quality", "evt_code", "evt_desc"]
        if include_serno:
            pivot_index.append("serno")
        df_pivoted = df.pivot(
            index=pivot_index,
            columns="suffix",
            values="value",
        ).reset_index()
        return df_pivoted

    def get_per_nmi_dfs(
        self,
        split_days: bool = False,
        set_interval: Optional[int] = None,
        include_serno: bool = False,
    ) -> Generator[Tuple[str, pd.DataFrame], None, None]:
        df = self.get_pivot_data_frame(split_days, set_interval, include_serno)
        nmis = df.nmi.unique()
        for nmi in nmis:
            nmi_df = df[(df["nmi"] == nmi)]
            del nmi_df["nmi"]
            yield nmi, nmi_df


def flatten_list(l: List[list]) -> list:
    """takes a list of lists, l and returns a flat list"""
    return [v for inner_l in l for v in inner_l]


def read_nem_file(file_path: str, ignore_missing_header=False) -> NEMData:
    """Read in NEM file and return meter readings named tuple

    :param file_path: The NEM file to process
    :param ignore_missing_header: Whether to continue parsing if missing header.
                                  Will assume NEM12 format.
    :returns: The file that was created
    """

    nf = NEMFile(file_path, strict=ignore_missing_header)
    return nf.nem_data()


def parse_100_row(row: List[Any], file_name: str) -> HeaderRecord:
    """Parse header record (100)"""
    return HeaderRecord(
        row[1], parse_datetime(row[2]), row[3], row[4], file_name, False
    )


def parse_nem12_rows(nem_list: Iterable, file_name=None) -> NEMReadings:
    """Parse NEM row iterator and return meter readings named tuple"""
    # readings nested by NMI then channel
    readings: Dict[str, Dict[str, List[Reading]]] = {}
    # transactions nested by NMI then channel
    trans: Dict[str, Dict[str, list]] = {}
    nmi_d = None  # current NMI details block that readings apply to

    observed_900_record = False

    for row in nem_list:

        if not row:
            log.debug("Skipping empty row.")
            continue

        record_indicator = int(row[0])

        if record_indicator == 900:
            # Powercor NEM12 files can concatenate multiple files together
            # try to keep parsing anyway.
            if observed_900_record:
                log.warning("Found multiple end of data (900) rows. ")

            observed_900_record = True
            pass

        elif record_indicator == 200:
            try:
                nmi_details = parse_200_row(row)
            except ValueError:
                log.error("Error passing 200 row:")
                log.error(row)
                raise
            nmi_d = nmi_details

            if nmi_d.nmi not in readings:
                readings[nmi_d.nmi] = {}
            if nmi_d.nmi_suffix not in readings[nmi_d.nmi]:
                readings[nmi_d.nmi][nmi_d.nmi_suffix] = []
            if nmi_d.nmi not in trans:
                trans[nmi_d.nmi] = {}
            if nmi_d.nmi_suffix not in trans[nmi_d.nmi]:
                trans[nmi_d.nmi][nmi_d.nmi_suffix] = []

        elif record_indicator == 300:
            num_intervals = int(24 * 60 / nmi_d.interval_length)
            assert len(row) > 1, f"Invalid 300 Row in {file_name}"
            if len(row) < num_intervals + 2:
                record_date = row[1]
                msg = "Skipping 300 record for %s %s %s. "
                msg += "It does not have the expected %s intervals"
                log.error(msg, nmi_d.nmi, nmi_d.nmi_suffix, record_date, num_intervals)
                continue
            interval_record = parse_300_row(
                row, nmi_d.interval_length, nmi_d.uom, nmi_d.meter_serial_number
            )
            # don't flatten the list of interval readings at this stage,
            # as they may need to be adjusted by a 400 row
            readings[nmi_d.nmi][nmi_d.nmi_suffix].append(
                interval_record.interval_values
            )

        elif record_indicator == 400:
            event_record = parse_400_row(row)
            readings[nmi_d.nmi][nmi_d.nmi_suffix][-1] = update_reading_events(
                readings[nmi_d.nmi][nmi_d.nmi_suffix][-1], event_record
            )

        elif record_indicator == 500:
            b2b_details = parse_500_row(row)
            trans[nmi_d.nmi][nmi_d.nmi_suffix].append(b2b_details)

        else:
            log.warning(
                "Record indicator %s not supported and was skipped", record_indicator
            )

    for nmi in readings:
        for suffix in readings[nmi]:
            readings[nmi][suffix] = flatten_list(readings[nmi][suffix])

    if not observed_900_record:
        log.warning("Missing end of data (900) row.")

    return NEMReadings(readings=readings, transactions=trans)


def parse_nem13_rows(nem_list: Iterable) -> NEMReadings:
    """Parse NEM row iterator and return meter readings named tuple"""
    # readings nested by NMI then channel
    readings: Dict[str, Dict[str, List[Reading]]] = {}
    # transactions nested by NMI then channel
    trans: Dict[str, Dict[str, list]] = {}
    nmi_d = None  # current NMI details block that readings apply to

    for row in nem_list:
        record_indicator = int(row[0])

        if record_indicator == 900:
            for nmi in readings:
                for suffix in readings[nmi]:
                    readings[nmi][suffix] = flatten_list(readings[nmi][suffix])
            break  # End of file

        elif record_indicator == 550:
            b2b_details = parse_550_row(row)
            trans[nmi_d.nmi][nmi_d.nmi_suffix].append(b2b_details)

        elif record_indicator == 250:
            basic_data = parse_250_row(row)
            reading = calculate_manual_reading(basic_data)

            nmi_d = basic_data

            if basic_data.nmi not in readings:
                readings[nmi_d.nmi] = {}
            if nmi_d.nmi_suffix not in readings[nmi_d.nmi]:
                readings[nmi_d.nmi][nmi_d.nmi_suffix] = []
            if nmi_d.nmi not in trans:
                trans[nmi_d.nmi] = {}
            if nmi_d.nmi_suffix not in trans[nmi_d.nmi]:
                trans[nmi_d.nmi][nmi_d.nmi_suffix] = []

            readings[nmi_d.nmi][nmi_d.nmi_suffix].append([reading])

        else:
            log.warning(
                "Record indicator %s not supported and was skipped", record_indicator
            )
    return NEMReadings(readings=readings, transactions=trans)


def calculate_manual_reading(basic_data: BasicMeterData) -> Reading:
    """Calculate the interval between two manual readings"""
    t_start = basic_data.previous_register_read_datetime
    t_end = basic_data.current_register_read_datetime
    val_start = basic_data.previous_register_read
    val_end = basic_data.current_register_read
    value = basic_data.quantity

    meter_serial_number = basic_data.meter_serial_number
    uom = basic_data.uom
    quality_method = basic_data.current_quality_method

    return Reading(
        t_start,
        t_end,
        value,
        uom,
        meter_serial_number,
        quality_method,
        "",
        "",
        val_start,
        val_end,
    )


def parse_200_row(row: list) -> NmiDetails:
    """Parse NMI data details record (200)"""

    next_read = None  # Next scheduled read is an optional field
    if len(row) > 9:
        next_read = parse_datetime(row[9])

    return NmiDetails(
        row[1], row[2], row[3], row[4], row[5], row[6], row[7], int(row[8]), next_read
    )


def parse_250_row(row: list) -> BasicMeterData:
    """Parse basic meter data record (250)"""
    return BasicMeterData(
        row[1],
        row[2],
        row[3],
        row[4],
        row[5],
        row[6],
        row[7],
        float(row[8]),
        parse_datetime(row[9]),
        row[10],
        row[11],
        row[12],
        float(row[13]),
        parse_datetime(row[14]),
        row[15],
        row[16],
        row[17],
        float(row[18]),
        row[19],
        row[20],
        parse_datetime(row[21]),
        parse_datetime(row[22]),
    )


def nth(iterable, n, default=None):
    "Returns the nth item or a default value"
    return next(islice(iterable, n, None), default)


def parse_300_row(
    row: list, interval: int, uom: str, meter_serial_number: str
) -> IntervalRecord:
    """Interval data record (300)"""

    num_intervals = int(24 * 60 / interval)
    interval_date = parse_datetime(row[1])
    last_interval = 2 + num_intervals
    quality_method = row[last_interval]

    # Optional fields
    reason_code = nth(row, last_interval + 1, "")
    reason_desc = nth(row, last_interval + 2, "")
    update_datetime = parse_datetime(nth(row, last_interval + 3, None))
    msats_load_datatime = parse_datetime(nth(row, last_interval + 4, None))

    interval_values = parse_interval_records(
        row[2:last_interval],
        interval_date,
        interval,
        uom,
        quality_method,
        meter_serial_number,
        reason_code,
        reason_desc,
    )
    return IntervalRecord(
        interval_date,
        interval_values,
        quality_method,
        meter_serial_number,
        reason_code,
        reason_desc,
        update_datetime,
        msats_load_datatime,
    )


def parse_interval_records(
    interval_record,
    interval_date,
    interval,
    uom,
    quality_method,
    meter_serial_number,
    event_code: str = "",
    event_desc: str = "",
) -> List[Reading]:
    """Convert interval values into tuples with datetime"""
    interval_delta = timedelta(minutes=interval)
    return [
        Reading(
            t_start=interval_date + (i * interval_delta),
            t_end=interval_date + (i * interval_delta) + interval_delta,
            read_value=parse_reading(val),
            uom=uom,
            quality_method=quality_method,
            meter_serial_number=meter_serial_number,
            event_code=event_code,  # This may get changed later by a 400 row
            event_desc=event_desc,  # This may get changed later by a 400 row
            val_start=None,
            val_end=None,  # No before and after readings for intervals
        )
        for i, val in enumerate(interval_record)
    ]


def parse_reading(val: str) -> Optional[float]:
    """Convert reading value to float (if possible)"""
    if val == "":
        return None
    try:
        return float(val)
    except ValueError:
        log.warning('Reading of "%s" is not a number', val)
        return None  # Not a valid reading


def parse_400_row(row: list) -> tuple:
    """Interval event record (400)"""

    return EventRecord(int(row[1]), int(row[2]), row[3], row[4], row[5])


def update_reading_events(readings, event_record):
    """Updates readings from a 300 row to reflect any events found in a
    subsequent 400 row
    """
    # event intervals are 1-indexed
    for i in range(event_record.start_interval - 1, event_record.end_interval):
        readings[i] = Reading(
            t_start=readings[i].t_start,
            t_end=readings[i].t_end,
            read_value=readings[i].read_value,
            uom=readings[i].uom,
            meter_serial_number=readings[i].meter_serial_number,
            quality_method=event_record.quality_method,
            event_code=event_record.reason_code,
            event_desc=event_record.reason_description,
            val_start=readings[i].val_start,
            val_end=readings[i].val_end,
        )
    return readings


def parse_500_row(row: list) -> tuple:
    """Parse B2B details record"""
    return B2BDetails12(row[1], row[2], row[3], row[4])


def parse_550_row(row: list) -> tuple:
    """Parse B2B details record"""
    return B2BDetails13(row[1], row[2], row[3], row[4])


def parse_datetime(record: str) -> Optional[datetime]:
    """Parse a datetime string into a python datetime object"""
    # NEM defines Date8, DateTime12 and DateTime14
    format_strings = {8: "%Y%m%d", 12: "%Y%m%d%H%M", 14: "%Y%m%d%H%M%S"}

    if record == "" or record is None:
        return None

    try:
        timestamp = datetime.strptime(
            record.strip(), format_strings[len(record.strip())]
        )
    except (ValueError, KeyError):
        log.debug(f"Malformed date '{record}' ")
        return None

    return timestamp
